﻿using System;
using System.Diagnostics;
using System.Globalization;
using System.Runtime.InteropServices;
using System.ComponentModel.Design;
using Microsoft.Win32;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.OLE.Interop;
using Microsoft.VisualStudio.Shell;
using ScaffoldIt.Commands;

namespace ScaffoldIt
{
    /// <summary>
    /// This is the class that implements the package exposed by this assembly.
    ///
    /// The minimum requirement for a class to be considered a valid package for Visual Studio
    /// is to implement the IVsPackage interface and register itself with the shell.
    /// This package uses the helper classes defined inside the Managed Package Framework (MPF)
    /// to do it: it derives from the Package class that provides the implementation of the 
    /// IVsPackage interface and uses the registration attributes defined in the framework to 
    /// register itself and its components with the shell.
    /// </summary>
    // This attribute tells the PkgDef creation utility (CreatePkgDef.exe) that this class is
    // a package.
    [PackageRegistration(UseManagedResourcesOnly = true)]
    // This attribute is used to register the informations needed to show the this package
    // in the Help/About dialog of Visual Studio.
    [InstalledProductRegistration("#110", "#112", "1.0", IconResourceID = 400)]
    // This attribute is needed to let the shell know that this package exposes some menus.
    [ProvideMenuResource("Menus.ctmenu", 1)]
    [ProvideAutoLoad(UIContextGuids80.SolutionExists)]
    [ProvideToolWindow(typeof(ScaffoldItToolWindow))]
    [Guid(GuidList.guidScaffoldItPkgString)]
    public sealed class ScaffoldItPackage : Package
    {
        private CommandSet _commandSet;
        
        /// <summary>
        /// Default constructor of the package.
        /// Inside this method you can place any initialization code that does not require 
        /// any Visual Studio service because at this point the package object is created but 
        /// not sited yet inside Visual Studio environment. The place to do all the other 
        /// initialization is the Initialize method.
        /// </summary>
        public ScaffoldItPackage()
        {
            Trace.WriteLine(string.Format(CultureInfo.CurrentCulture, "Entering constructor for: {0}", this.ToString()));
        }

        /// <summary>
        /// This function is called when the user clicks the menu item that shows the 
        /// tool window. See the Initialize method to see how the menu item is associated to 
        /// this function using the OleMenuCommandService service and the MenuCommand class.
        /// </summary>
        private void ShowToolWindow(object sender, EventArgs e)
        {
            // Get the instance number 0 of this tool window. This window is single instance so this instance
            // is actually the only one.
            // The last flag is set to true so that if the tool window does not exists it will be created.
            ShowToolWindow();
        }

        public ScaffoldItToolWindow ShowToolWindow()
        {
            ScaffoldItToolWindow window = (ScaffoldItToolWindow)this.FindToolWindow(typeof(ScaffoldItToolWindow), 0, true);
            if ((null == window) || (null == window.Frame))
            {
                throw new NotSupportedException(Resources.CanNotCreateWindow);
            }
            IVsWindowFrame windowFrame = (IVsWindowFrame)window.Frame;
            Microsoft.VisualStudio.ErrorHandler.ThrowOnFailure(windowFrame.Show());
            return window;
        }

        /////////////////////////////////////////////////////////////////////////////
        // Overriden Package Implementation
        #region Package Members

        private EnvDTE.SolutionEvents solevents;

        /// <summary>
        /// Initialization of the package; this method is called right after the package is sited, so this is the place
        /// where you can put all the initilaization code that rely on services provided by VisualStudio.
        /// </summary>
        protected override void Initialize()
        {
            Trace.WriteLine (string.Format(CultureInfo.CurrentCulture, "Entering Initialize() of: {0}", this.ToString()));
            base.Initialize();
            
            try
            {
                var dte = (EnvDTE80.DTE2)Microsoft.VisualStudio.Shell.Package.GetGlobalService(typeof(Microsoft.VisualStudio.Shell.Interop.SDTE));
                solevents = dte.Events.SolutionEvents;
                solevents.AfterClosing += new EnvDTE._dispSolutionEvents_AfterClosingEventHandler(SolutionEvents_AfterClosing);
                solevents.Opened += new EnvDTE._dispSolutionEvents_OpenedEventHandler(SolutionEvents_Opened);
                _commandSet = new CommandSet(this);
                var mcs = (OleMenuCommandService)GetService(typeof(IMenuCommandService));
                _commandSet.InitializeCommands(mcs);

                // Create the command for the tool window
                CommandID toolwndCommandID = new CommandID(GuidList.guidScaffoldItCmdSet, (int)PkgCmdIDList.cmdShowScaffoldIt);
                MenuCommand menuToolWin = new MenuCommand(ShowToolWindow, toolwndCommandID);
                mcs.AddCommand(menuToolWin);
                SolutionEvents_Opened();
            }
            catch (Exception ex)
            {
                WriteToActivityLog(__ACTIVITYLOG_ENTRYTYPE.ALE_ERROR, "Expect package throw an error\r\n" + ex.ToString());
                throw;
            }
        }

        void SolutionEvents_Opened()
        {
            var wnd = (ScaffoldItToolWindow)this.FindToolWindow(typeof(ScaffoldItToolWindow), 0, true);
            if (wnd != null)
            {
                wnd.RefreshWindow();
            }
        }

        void SolutionEvents_AfterClosing()
        {
            var wnd = (ScaffoldItToolWindow)this.FindToolWindow(typeof(ScaffoldItToolWindow), 0, true);
            if (wnd != null)
            {
                wnd.CloseConfig();
            }
        }
        #endregion

        private void WriteToActivityLog(__ACTIVITYLOG_ENTRYTYPE type, string msg)
        {
            IVsActivityLog log = GetService(typeof(SVsActivityLog)) as IVsActivityLog;
            if (log == null) return;

            int hr = log.LogEntry((UInt32)type, this.ToString(), msg);
        }
    }
}
